/*
 * i4storage.cpp --
 *
 *	This file contains the implementation of the e4_StorageImpl class
 *	defined in e4graph.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graphimpl.h"

static bool initialized = false;
static e4_HashTable *activeStorages = NULL;

/*
 * This static function is the only public way to get an e4_StorageImpl.
 */

e4_StorageImpl *
e4_StorageImpl::GetStorage(const char *name, const char *drivername)
{
    e4_HashEntry *hPtr;
    e4_StorageImpl *s;
    e4_StorageConstructor scp;
    int isnew;

    if (!initialized) {
	initialized = true;
	e4_InitializeStorageRegistry();
	activeStorages = e4_NewHashTable(E4_STRING_KEYS);
    }
    hPtr = E4_CREATEHASHENTRY(activeStorages, name, &isnew);
    if (!isnew) {
	return (e4_StorageImpl *) E4_GETHASHVALUE(hPtr);
    }
    scp = e4_GetStorageConstructor(drivername);
    if (scp == NULL) {
	s = NULL;
    } else {
	s = (*scp)(name);
    }
    if (s == NULL) {
	e4_DeleteHashEntry(hPtr);
    } else {
	E4_SETHASHVALUE(hPtr, s);
    }
    return s;
}

/*
 * This constructor opens a storage in write mode.
 */

e4_StorageImpl::e4_StorageImpl(const char *fname, const char *dnm) 
    : e4_RefCounter()
{
    drivername = new char[strlen(dnm) + 1];
    strcpy(drivername, dnm);

    name = new char[strlen(fname) + 1];
    strcpy(name, fname);

    commit = true;
    dirty = false;
    locked = false;

    InitializeNameHash();
    InitializeNodeHash();
    InitializeVertexHash();
    InitializeCallbackHash();
}

/*
 * The destructor is virtual so it is called before destructors of derived
 * classes to clean up the state belonging to the e4_StorageImpl base class.
 */

e4_StorageImpl::~e4_StorageImpl()
{
    if (name != NULL) {
	/*
	 * We are supposed to delete the storage occupied by the name
	 * but because of bugs in some compilers we cannot delete it.
	 * This is because attempting to delete it with e.g. VC6 causes
	 * an assert at this point. We therefore just leak the string.
	 */
    }
}
    
/*
 * Helper functions to initialize and manage the caches.
 */

void
e4_StorageImpl::InitializeNameHash()
{
    nameHash = e4_NewHashTable(E4_STRING_KEYS);
}

void
e4_StorageImpl::InitializeNodeHash()
{
    activeNodes = e4_NewHashTable(E4_ONE_WORD_KEYS);
}

void
e4_StorageImpl::InitializeVertexHash()
{
    activeVertices = e4_NewHashTable(E4_ONE_WORD_KEYS);
}

void
e4_StorageImpl::InitializeCallbackHash()
{
    callbacks = e4_NewHashTable(3);
    callbacksPresent = 0;
}

/*
 * Retrieve an instance of e4_VertexImpl for a given vertex.
 */

e4_VertexImpl *
e4_StorageImpl::GetVertex(int i) const
{
    e4_VertexImpl *f;
    e4_HashEntry *hPtr;
    int isNew;

    if (!IsLegalVertexID(i) || (activeVertices == NULL)) {
	return NULL;
    }
    hPtr = E4_CREATEHASHENTRY(activeVertices, (const char *) i, &isNew);
    if (isNew) {
        f = new e4_VertexImpl((e4_StorageImpl *) this, i);
        E4_SETHASHVALUE(hPtr, f);
    } else {
        f = (e4_VertexImpl *) E4_GETHASHVALUE(hPtr);
    }
    return f;
}

/*
 * Store a node in the activeNodes hash table.
 */

void
e4_StorageImpl::StoreNode(int index, e4_NodeImpl *n) const
{
    e4_HashEntry *hPtr;
    int isnew;

    if (activeNodes == NULL) {
	return;
    }
    hPtr = E4_CREATEHASHENTRY(activeNodes, (const char *) index, &isnew);
    E4_SETHASHVALUE(hPtr, n);
}

/*
 * Remove a node from the activeNodes hash table. This method should only
 * be called from the e4_NodeImpl destructor, to avoid infinite recursion.
 */

void
e4_StorageImpl::ForgetNode(int index)
{
    e4_HashEntry *hPtr;
    e4_NodeImpl *n;

    if (activeNodes == NULL) {
	return;
    }
    hPtr = E4_FINDHASHENTRY(activeNodes, (const char *) index);
    if (hPtr != NULL) {
        n = (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
        n->SetStorage(NULL);
        e4_DeleteHashEntry(hPtr);
    }
}

/*
 * Find a node in the activeNodes hash table given its index, if it is
 * in-memory. If not, check whether the index is legal and create it
 * if it is.
 */

e4_NodeImpl *
e4_StorageImpl::FindNode(int index) const
{
    if (IsLegalNodeID(index)) {
	return FindOrCreateNode(index);
    }
    return NULL;
}

/*
 * Find a node in the activeNodes hash table given its index, if it is
 * in-memory. If not, create it. This method does not check whether
 * the index is legal.
 */

e4_NodeImpl *
e4_StorageImpl::FindOrCreateNode(int index) const
{
    e4_HashEntry *hPtr;
    e4_NodeImpl *n;

    if (activeNodes == NULL) {
	return NULL;
    }
    hPtr = E4_FINDHASHENTRY(activeNodes, (const char *) index);
    if (hPtr != NULL) {
	return (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
    }
    n = new e4_NodeImpl((e4_StorageImpl *) this, index);
    StoreNode(index, n);
    
    return n;
}

/*
 * Find a node in the activeNodes hash table given its index, if it is
 * in-memory. If not, create it and store it in the activeNodes table.
 * Also return an indication whether the node was newly created.
 */

e4_NodeImpl *
e4_StorageImpl::FindOrCreateNode(int index, bool *isNew) const
{
    e4_HashEntry *hPtr;
    e4_NodeImpl *n;

    if (activeNodes == NULL) {
	return NULL;
    }
    hPtr = E4_FINDHASHENTRY(activeNodes, (const char *) index);
    if (hPtr != NULL) {
        *isNew = false;
	return (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
    }
    *isNew = true;
    n = new e4_NodeImpl((e4_StorageImpl *) this, index);
    StoreNode(index, n);
    
    return n;
}

/*
 * Store a vertex in the activeVertices hash table.
 */

void
e4_StorageImpl::StoreVertex(int vertexID, e4_VertexImpl *f) const
{
    e4_HashEntry *hPtr;
    int isNew;

    if (activeVertices == NULL) {
	return;
    }
    hPtr = E4_CREATEHASHENTRY(activeVertices, (const char *) vertexID, &isNew);
    if (!isNew) {

        /*
         * Should abort here.
         */
    }
    E4_SETHASHVALUE(hPtr, f);
}

/*
 * Remove an e4_VertexImpl from the activeVertices hash table.
 */

void
e4_StorageImpl::ForgetVertex(int vertexID)
{
    e4_HashEntry *hPtr;
    e4_VertexImpl *f;

    if (activeVertices == NULL) {
	return;
    }
    hPtr = E4_FINDHASHENTRY(activeVertices, (const char *) vertexID);
    if (hPtr != NULL) {
	f = (e4_VertexImpl *) E4_GETHASHVALUE(hPtr);
	f->SetStorage(NULL);
        e4_DeleteHashEntry(hPtr);
    }
}

/*
 * Is the node with this ID referenced by the user of e4Graph?
 */

bool
e4_StorageImpl::IsReferencedNode(int nodeID) const
{
    if (activeNodes == NULL) {
	return false;
    }
    if (E4_FINDHASHENTRY(activeNodes, (const char *) nodeID) == NULL) {
	return false;
    }
    return true;
}

/*
 * Iterate over all nodes referenced by the user program.
 */

int
e4_StorageImpl::FirstReferencedNodeID(e4_HashSearch *searchP)
{
    e4_HashEntry *ePtr;

    if (activeNodes == NULL) {
	return E4_NEXTNONE;
    }
    ePtr = e4_FirstHashEntry(activeNodes, searchP);
    if (ePtr == NULL) {
	return E4_NEXTNONE;
    }
    return (int) E4_GETHASHKEY(activeNodes, ePtr);
}

int
e4_StorageImpl::NextReferencedNodeID(e4_HashSearch *searchP)
{
    e4_HashEntry *ePtr;

    if (activeNodes == NULL) {
	return E4_NEXTNONE;
    }
    ePtr = e4_NextHashEntry(searchP);
    if (ePtr == NULL) {
	return E4_NEXTNONE;
    }
    return (int) E4_GETHASHKEY(activeNodes, ePtr);
}

/*
 * Is the vertex with this ID referenced by the user of e4Graph?
 */

bool
e4_StorageImpl::IsReferencedVertex(int vertexID) const
{
    if (activeVertices == NULL) {
	return false;
    }
    if (E4_FINDHASHENTRY(activeVertices, (const char *) vertexID) == NULL) {
	return false;
    }
    return true;
}

/*
 * Iterate over all vertices referenced by the user program.
 */

int
e4_StorageImpl::FirstReferencedVertexID(e4_HashSearch *searchP)
{
    e4_HashEntry *ePtr;

    if (activeVertices == NULL) {
	return E4_NEXTNONE;
    }
    ePtr = e4_FirstHashEntry(activeVertices, searchP);
    if (ePtr == NULL) {
	return E4_NEXTNONE;
    }
    return (int) E4_GETHASHKEY(activeVertices, ePtr);
}

int
e4_StorageImpl::NextReferencedVertexID(e4_HashSearch *searchP)
{
    e4_HashEntry *ePtr;

    if (activeVertices == NULL) {
	return E4_NEXTNONE;
    }
    ePtr = e4_NextHashEntry(searchP);
    if (ePtr == NULL) {
	return E4_NEXTNONE;
    }
    return (int) E4_GETHASHKEY(activeVertices, ePtr);
}

/*
 * Add a name to the name hash table.
 */

void
e4_StorageImpl::AddNameToNameHash(const char *nm, int idx) const
{
    e4_HashEntry *hPtr;
    int isnew;

    if (nameHash == NULL) {
	return;
    }
    hPtr = E4_CREATEHASHENTRY(nameHash, nm, &isnew);
    E4_SETHASHVALUE(hPtr, idx);
}

/*
 * Return a name given its nameID.
 */

const char *
e4_StorageImpl::NameFromNameID(int nameID) const
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;

    if (nameHash == NULL) {
	return NULL;
    }

    /*
     * Search through the hash table until we find the hash entry with
     * the nameID as its value. If found, return the key which is the
     * name that hashes to this nameID.
     */

    for (hPtr = e4_FirstHashEntry(nameHash, &search);
	 hPtr != NULL;
	 hPtr = e4_NextHashEntry(&search)) {
	if ((int) E4_GETHASHVALUE(hPtr) == nameID) {
	    return (const char *) E4_GETHASHKEY(nameHash, hPtr);
	}
    }

    /*
     * Not found, return NULL.
     */

    return NULL;
}

/*
 * Get the e4_NodeImpl for the root node of this storage:
 */

e4_NodeImpl *
e4_StorageImpl::GetRootNode()
{
    return FindOrCreateNode(GetRootNodeID());
}

/*
 * Set the root node for this storage. If the old root is detached
 * and not exported, then recycle memory.
 */

bool
e4_StorageImpl::SetRootNode(e4_NodeImpl *nrp)
{
    int ornID, nrnID;

    /*
     * Sanity checks.
     */

    if ((nrp == NULL) || (!nrp->IsValid())) {
	return false;
    }
    nrnID = nrp->GetUniqueID();
   
    /*
     * Remember the old root node ID.
     */

    ornID = GetRootNodeID();

    /*
     * Set the new root node ID.
     */

    SetRootNodeID(nrnID);

    /*
     * If the old root node is detached and not exported, recycle memory.
     */

    if (IsDetachedNodeID(ornID) && !IsReferencedNode(ornID)) {
	if (!IsGCDeferred()) {
	    DoGC();
	} else {
	    SetNeedsGC(true);
	}
    }

    return true;
}

/*
 * Add a name to the persistent storage and also intern it in the
 * hash table.
 */

int
e4_StorageImpl::InternName(const char *nm) const
{
    e4_HashEntry *hPtr;
    int isnew, idx;
    
    if (nameHash == NULL) {
	return E4_NEXTNONE;
    }

    /*
     * See if we already have this name in the hash table; if so, just
     * return the hash value.
     */

    hPtr = E4_FINDHASHENTRY(nameHash, nm);
    if (hPtr != NULL) {
	return (int) E4_GETHASHVALUE(hPtr);
    }

    /*
     * No, we need to add the name to hash table.
     */

    idx = ((e4_StorageImpl *) this)->AddName(nm);
    hPtr = E4_CREATEHASHENTRY(nameHash, nm, &isnew);
    E4_SETHASHVALUE(hPtr, idx);

    return idx;
}

/*
 * This method is called to clean up all references to this e4_StorageImpl
 * from other instances.
 */

void
e4_StorageImpl::CleanUp()
{
    /*
     * If commit is needed and requested, do it now before the state
     * is destroyed.
     */
    
    if (AutoCommit() && IsDirty()) {
	Commit();
    }

    HashCleanup();
}

/*
 * Helper method to clean up the hash tables for various exported entities.
 */

void
e4_StorageImpl::HashCleanup()
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;
    e4_NodeImpl *node;
    e4_VertexImpl *vertex;

    /*
     * Clean up my own state, consisting of the hash tables for
     * caching in-core instances of nodes and vertices.
     */

    if (activeNodes != NULL) {
	for (hPtr = e4_FirstHashEntry(activeNodes, &search);
	     hPtr != NULL;
	     hPtr = e4_FirstHashEntry(activeNodes, &search)) {
	    node = (e4_NodeImpl *) E4_GETHASHVALUE(hPtr);
	    node->SetStorage(NULL);
	    e4_DeleteHashEntry(hPtr);
	}
    }
    if (activeVertices != NULL) {
	for (hPtr = e4_FirstHashEntry(activeVertices, &search);
	     hPtr != NULL;
	     hPtr = e4_FirstHashEntry(activeVertices, &search)) {
	    vertex = (e4_VertexImpl *) E4_GETHASHVALUE(hPtr);
	    vertex->SetStorage(NULL);
	    e4_DeleteHashEntry(hPtr);
	}
    }
    if (nameHash != NULL) {
	for (hPtr = e4_FirstHashEntry(nameHash, &search);
	     hPtr != NULL;
	     hPtr = e4_FirstHashEntry(nameHash, &search)) {
	    e4_DeleteHashEntry(hPtr);
	}
    }
}

/*
 * This method is called when the reference count goes to (or below) zero.
 */

void
e4_StorageImpl::NotReferenced()
{
    e4_HashEntry *hPtr;

    CleanUp();

    /*
     * Delete the hash tables associated with this instance so that
     * we do not leak memory.
     */

    e4_DeleteHashTable(activeNodes);
    e4_DeleteHashTable(activeVertices);
    e4_DeleteHashTable(nameHash);
    e4_DeleteHashTable(callbacks);
    activeNodes = NULL;
    activeVertices = NULL;
    nameHash = NULL;
    callbacks = NULL;

    /*
     * Cause a last GC to clean up detached entities, before we commit.
     */

    DoGC();

    /*
     * Remove this e4_StorageImpl from the global cache of 
     * e4_StorageImpl instances.
     */
    
    hPtr = E4_FINDHASHENTRY(activeStorages, name);
    if (hPtr == NULL) {
	return;
    }
    e4_DeleteHashEntry(hPtr);

    if (name != NULL) {
	delete[] name;
    }
    if (drivername != NULL) {
	delete[] drivername;
    }

    /*
     * Call the destructor which is most likely going to be overridden by
     * the derived class.
     */
    
    delete this;
}

/*
 * Retrieve the first e4_StorageImpl.
 */

e4_StorageImpl *
e4_StorageImpl::GetFirstStorageImpl()
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;

    if (activeStorages == NULL) {
	return NULL;
    }
    hPtr = e4_FirstHashEntry(activeStorages, &search);
    if (hPtr == NULL) {
	return NULL;
    }
    return (e4_StorageImpl *) E4_GETHASHVALUE(hPtr);
}

/*
 * Retrieve the next e4_StorageImpl after a given one.
 */

e4_StorageImpl *
e4_StorageImpl::GetNextStorageImpl(e4_StorageImpl *cur)
{
    e4_HashEntry *hPtr;
    e4_HashSearch search;

    if (cur == NULL) {
	return NULL;
    }
    if (activeStorages == NULL) {
	return NULL;
    }
    for (hPtr = e4_FirstHashEntry(activeStorages, &search);
	 hPtr != NULL;
	 hPtr = e4_NextHashEntry(&search)) {
	if (cur == (e4_StorageImpl *) E4_GETHASHVALUE(hPtr)) {
	    hPtr = e4_NextHashEntry(&search);
	    if (hPtr == NULL) {
		return NULL;
	    }
	    return (e4_StorageImpl *) E4_GETHASHVALUE(hPtr);
	}
    }

    /*
     * Current one not found, return NULL.
     */

    return NULL;
}

/*
 * Add a callback.
 */

void
e4_StorageImpl::AddCallback(e4_CallbackKind kind, void *fn, void *clientData)
{
    int key[3];
    e4_HashEntry *ePtr;
    int isnew;

    key[0] = (int) kind;
    key[1] = (int) fn;
    key[2] = (int) clientData;

    ePtr = E4_CREATEHASHENTRY(callbacks, (char *) key, &isnew);
    E4_SETHASHVALUE(ePtr, (void *) NULL);

    callbacksPresent |= (1 << (int) kind);
}

/*
 * Remove a callback.
 */

void
e4_StorageImpl::DelCallback(e4_CallbackKind kind, void *fn, void *clientData)
{
    int key[3], *keyPtr;
    e4_HashEntry *ePtr;
    e4_HashSearch search;

    key[0] = (int) kind;
    key[1] = (int) fn;
    key[2] = (int) clientData;

    ePtr = E4_FINDHASHENTRY(callbacks, (char *) key);
    if (ePtr != NULL) {
	e4_DeleteHashEntry(ePtr);
    }

    /*
     * Now see if there are any other callbacks of this kind left. If not,
     * then remove the notation that this kind of callback is present for
     * this storage.
     */

    for (ePtr = e4_FirstHashEntry(callbacks, &search);
	 ePtr != NULL;
	 ePtr = e4_NextHashEntry(&search)) {
	keyPtr = (int *) E4_GETHASHKEY(callbacks, ePtr);
	if (keyPtr[0] == (int) kind) {
	    return;
	}
    }

    /*
     * Did not find any, so remove the notation.
     */

    callbacksPresent &=(~(1 << (int) kind));
}

/*
 * Cause callbacks to be called for a specified kind of event on the
 * given reference.
 */

void
e4_StorageImpl::CallbackEvent(e4_CallbackKind kind, e4_RefCounter *ref)
{
    e4_NodeAddCallbackFunction nacbfn;
    e4_NodeDetCallbackFunction ntcbfn;
    e4_VertexAddCallbackFunction vacbfn;
    e4_VertexDetCallbackFunction vtcbfn;
    e4_VertexModCallbackFunction vmcbfn;
    e4_HashEntry *ePtr;
    e4_HashSearch search;
    int *key;
    bool oldlocking;

    /*
     * Lock the storage so no changes are allowed.
     */

    oldlocking = SetLocking(true);

    switch (kind) {
    case E4_CBADDNODE:
	{
	    e4_Node n1((e4_NodeImpl *) ref);

	    if (callbacks == NULL) {
		break;
	    }
	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBADDNODE) {
		    nacbfn = (e4_NodeAddCallbackFunction) key[1];
		    (nacbfn)((void *) key[2], n1);
		}
	    }
	}
	break;
    case E4_CBDETNODE:
	{
	    e4_Node n3((e4_NodeImpl *) ref);

	    if (callbacks == NULL) {
		break;
	    }
	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBDETNODE) {
		    ntcbfn = (e4_NodeDetCallbackFunction) key[1];
		    (ntcbfn)((void *) key[2], n3);
		}
	    }
	}
	break;
    case E4_CBADDVERTEX:
	{
	    e4_Vertex v1((e4_VertexImpl *) ref);

	    if (callbacks == NULL) {
		break;
	    }
	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBADDVERTEX) {
		    vacbfn = (e4_VertexAddCallbackFunction) key[1];
		    (vacbfn)((void *) key[2], v1);
		}
	    }
	}
	break;
    case E4_CBDETVERTEX:
	{
	    e4_Vertex v3((e4_VertexImpl *) ref);

	    if (callbacks == NULL) {
		break;
	    }
	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBDETVERTEX) {
		    vtcbfn = (e4_VertexDetCallbackFunction) key[1];
		    (vtcbfn)((void *) key[2], v3);
		}
	    }
	}
	break;
    case E4_CBMODVERTEX:
        {
	    e4_Vertex v4((e4_VertexImpl *) ref);

	    if (callbacks == NULL) {
		break;
	    }
	    for (ePtr = e4_FirstHashEntry(callbacks, &search);
		 ePtr != NULL;
		 ePtr = e4_NextHashEntry(&search)) {
		key = (int *) E4_GETHASHKEY(callbacks, ePtr);
		if (key[0] == E4_CBMODVERTEX) {
		    vmcbfn = (e4_VertexModCallbackFunction) key[1];
		    (vmcbfn)((void *) key[2], v4);
		}
	    }
	}
	break;
    }

    (void) SetLocking(oldlocking);
}

/*
 * Lock or unlock the storage.
 */

bool
e4_StorageImpl::SetLocking(bool lockvalue) 
{
    bool oldvalue = locked;

    locked = lockvalue;
    return oldvalue;
}

/*
 * Get the node ID of a given node.
 */

int
e4_StorageImpl::GetNodeID(e4_Node n) const
{
    if (n.impl == NULL) {
	return E4_NEXTNONE;
    }
    return ((e4_NodeImpl *) n.impl)->nodeID;
}

/*
 * Create a detached node (not attached to any vertex):
 */

e4_NodeImpl *
e4_StorageImpl::CreateDetachedNode()
{
    return FindOrCreateNode(ReserveNodeID());
}

/*
 * Create a detached vertex (not attached to a node).
 */

e4_VertexImpl *
e4_StorageImpl::CreateDetachedVertex(const char *nm, e4_NodeImpl *nip)
{
    int vertexID;
    int nodeID;
    e4_VertexImpl *vip;

    vertexID = ReserveVertexID(nm);
    if (vertexID == E4_VERTEXNOTFOUND) {
	return NULL;
    }

    /*
     * Must export vertex to user program before setting its value,
     * otherwise the SeVertexByIndex will fail.
     */

    vip = new e4_VertexImpl(this, vertexID);
    StoreVertex(vertexID, vip);

    nodeID = nip->GetUniqueID();
    if (!SetVertexByIndexToNode(vertexID, nodeID)) {
	return NULL;
    }
    return vip;
}

e4_VertexImpl *
e4_StorageImpl::CreateDetachedVertex(const char *nm, int i)
{
    int vertexID;
    e4_VertexImpl *vip;

    vertexID = ReserveVertexID(nm);
    if (vertexID == E4_VERTEXNOTFOUND) {
	return NULL;
    }

    /*
     * Must export vertex to user program before setting its value,
     * otherwise the SeVertexByIndex will fail.
     */

    vip = new e4_VertexImpl(this, vertexID);
    StoreVertex(vertexID, vip);

    if (!SetVertexByIndex(vertexID, i)) {
	return NULL;
    }
    return vip;
}

e4_VertexImpl *
e4_StorageImpl::CreateDetachedVertex(const char *nm, double d)
{
    int vertexID;
    e4_VertexImpl *vip;

    vertexID = ReserveVertexID(nm);
    if (vertexID == E4_VERTEXNOTFOUND) {
	return NULL;
    }

    /*
     * Must export vertex to user program before setting its value,
     * otherwise the SeVertexByIndex will fail.
     */

    vip = new e4_VertexImpl(this, vertexID);
    StoreVertex(vertexID, vip);

    if (!SetVertexByIndex(vertexID, d)) {
	return NULL;
    }
    return vip;
}

e4_VertexImpl *
e4_StorageImpl::CreateDetachedVertex(const char *nm, const char *s)
{
    int vertexID;
    e4_VertexImpl *vip;

    vertexID = ReserveVertexID(nm);
    if (vertexID == E4_VERTEXNOTFOUND) {
	return NULL;
    }

    /*
     * Must export vertex to user program before setting its value,
     * otherwise the SeVertexByIndex will fail.
     */

    vip = new e4_VertexImpl(this, vertexID);
    StoreVertex(vertexID, vip);

    if (!SetVertexByIndex(vertexID, s)) {
	return NULL;
    }
    return vip;
}

e4_VertexImpl *
e4_StorageImpl::CreateDetachedVertex(const char *nm, const void *b, int nb)
{
    int vertexID;
    e4_VertexImpl *vip;

    vertexID = ReserveVertexID(nm);
    if (vertexID == E4_VERTEXNOTFOUND) {
	return NULL;
    }

    /*
     * Must export vertex to user program before setting its value,
     * otherwise the SeVertexByIndex will fail.
     */

    vip = new e4_VertexImpl(this, vertexID);
    StoreVertex(vertexID, vip);

    if (!SetVertexByIndex(vertexID, b, nb)) {
	return NULL;
    }
    return vip;
}

/*
 * Cause deferred node and vertex addition events.
 */

void
e4_StorageImpl::CauseDeferredNodeAdditionEvent(e4_NodeImpl *nip)
{
    if (HasCallbacks(E4_CBADDNODE)) {
	CallbackEvent(E4_CBADDNODE, nip);
    }
}

void
e4_StorageImpl::CauseDeferredVertexAdditionEvent(e4_VertexImpl *vip)
{
    if (HasCallbacks(E4_CBADDVERTEX)) {
	CallbackEvent(E4_CBADDVERTEX, vip);
    }
}
