/*
 * e4graphimpl.h --
 *
 *	This file contains the definition of the implementation part of the
 *	e4Graph library. It defines the following classes:
 *
 *	e4_RefCounter	-- A base class used to manage lifecycle.
 *	e4_StorageImpl	-- Storage independent implementation of tree
 *			   data storage.
 *	e4_NodeImpl	-- An implementation of a node.
 *	e4_VertexImpl	-- An implementation of a vertex.
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef	__E4_GRAPHIMPL_H__
#define	__E4_GRAPHIMPL_H__

#if (defined(__GNUC__)) || (!defined(_WIN32))
#include <unistd.h>
#endif

#include "e4graph.h"

/*
 * Public #define values.
 */

/*
 * The value of E4_NEXTNONE is used to indicate the end of a chain
 * of records.
 */

#define E4_NEXTNONE		-1

/*
 * Get the hash table definitions used internally:
 */

#include "e4hash.h"

/*
 * The e4_RefCounter class:
 *
 * This class is the base class for all of the e4 classes, and implements
 * simple refcount controlled lifecycle mgmt for objects. Objects start out
 * with a zero refcount, and when the refcount goes back to zero, a virtual
 * method NotReferenced is called. This allows the derived class to decide
 * what to do.
 */

class e4_DLL e4_RefCounter {
private:
    /*
     * This field contains the current number of outstanding references
     * to this object.
     */

    int refCount;

    /*
     * User level data attached to this instance.
     */

    void *userData;

protected:
    /*
     * Constructor:
     */

    e4_RefCounter() : refCount(0), userData(NULL) {}

    /*
     * This method, which must be supplied by derived classes, is called
     * each time the reference count goes back to zero.
     */

    virtual void NotReferenced() = 0;

public:

    /*
     * Reference count management:
     */

    inline void IncrRefCount() {
	refCount++;
#ifdef	DEBUG_REFCOUNT
	fprintf(stderr, "0x%x: refcount %d\n", this, refCount);
#endif
    }
    inline void DecrRefCount() {
	refCount--;
#ifdef	DEBUG_REFCOUNT
	fprintf(stderr, "0x%x: refcount %d\n", this, refCount);
#endif
	if (refCount <= 0) {
	     NotReferenced();
	}
    }

    /*
     * Returns the current reference count.
     */

    inline int RefCount() const
    {
	return refCount;
    }

    /*
     * These operations manipulate user-level data attached to each
     * instance.
     */

    inline void SetTransientUserData(void *d) {
	userData = d;
    }
    inline void *GetTransientUserData() {
	return userData;
    }

    /*
     * Is this instance valid? This method *must* be overridden
     * by sub-classes.
     */

    virtual bool IsValid() const = 0;
};

/*
 * The e4_StorageImpl class:
 *
 * NOTE: Because most of the operations in this class use c4_Property
 * instances which are declared extern but not exported, this class cannot
 * have any inline operations (or at least it cannot have any inline
 * operations that use these instances...)
 */

class e4_DLL e4_StorageImpl : public e4_RefCounter {
private:
    bool commit;			/* Should the storage be committed
					 * when the e4_StorageImpl instance is
					 * destroyed? */
    bool dirty;				/* Are there pending uncommitted
					 * changes? */
    e4_HashTable *activeNodes;		/* Hash table for nodes that
                                         * have been loaded in-memory
					 * to allow returning the same
                                         * pointer on multiple requests
					 * for a specific node. */
    e4_HashTable *activeVertices;	/* Hash table for vertices that
                                         * habe been loaded in-memory
                                         * to allow returning the same
                                         * pointer on multiple requests
                                         * for a specific vertex. */
    e4_HashTable *nameHash;		/* Hash table to map from
                                         * vertex names to indices in
                                         * the "names" view. Used to
                                         * allow use of integer indices
                                         * for vertex names instead of
                                         * the more inefficient string
                                         * values. */

    e4_HashTable *callbacks;		/* Hash table for callbacks defined
					 * for this storage. */

    int callbacksPresent;		/* Bitmap for notating that a certain
					 * kind of callback is present. This
					 * is used to efficiently decide if
					 * we need to invoke the callback
					 * mechanism for a given kind of
					 * event. */
    bool locked;			/* Usually false, set to true when
					 * changes to the storage are not
					 * allowed (e.g. when a deletion is
					 * in progress). */

protected:
    char *drivername;			/* The name of the storage driver
					 * being used by this storage. */
    char *name;				/* Unique name for storage. */

private:
    e4_StorageImpl();			/* The default constructor is private
					 * to prevent it from being called. */
    e4_StorageImpl(const e4_StorageImpl &);
					/* Copy constructor cannot be
					 * called either. */
    void operator= (const e4_StorageImpl *);
					/* Nor assignment. */

    /*
     * These operations are for use by e4_StorageImpl only:
     */

    void InitializeNameHash();
    void InitializeNodeHash();
    void InitializeVertexHash();
    void InitializeCallbackHash();

public:

    /*
     * Use this static function to get an e4_StorageImpl:
     */

    static e4_StorageImpl *GetStorage(const char *name, const char *drivernm);
    
    /*
     * Set and get the commit policy.
     */

    inline void AutoCommit(bool on) {
	commit = on;
    }
    inline bool AutoCommit() {
        return commit;
    }

    /*
     * Is this storage dirty? Mark the storage as dirty.
     */

    inline bool IsDirty() {return dirty;};
    inline void MarkDirty() {dirty = true;};

    /*
     * Is the storage temporarily locked while a deletion is in progress?
     */

    inline bool IsLocked() {return locked;};

    /*
     * Get the name and storage driver ID for this storage.
     */

    inline const char *GetName() const {return name;}
    inline const char *GetDriver() const {return drivername;}

    /*
     * Get and set the root node.
     */

    e4_NodeImpl *GetRootNode();
    bool SetRootNode(e4_NodeImpl *nrp);

    /*
     * Enter a name into the persistent storage and also into storage-wide
     * hash table, or retrieve its index from the hash table if it is
     * already interned.
     */

    int InternName(const char *nm) const;

    /*
     * Add a callback function.
     */

    void AddCallback(e4_CallbackKind kind, void *fn, void *clientData);

    /*
     * Remove a callback function.
     */

    void DelCallback(e4_CallbackKind kind, void *fn, void *clientData);

    /*
     * Cause callbacks to be called for a specified kind of event on the
     * given reference.
     */

    void CallbackEvent(e4_CallbackKind kind, e4_RefCounter *ref);

    /*
     * Does this storage have a callback of the given kind?
     */

    inline bool HasCallbacks(e4_CallbackKind kind) {
	return ((callbacksPresent & (1 << (int) kind)) == (1 << (int) kind))
	    ? true
	    : false;
    }

    /*
     * Create detached entities within this storage.
     */

    e4_NodeImpl   *CreateDetachedNode();
    e4_VertexImpl *CreateDetachedVertex(const char *nm, e4_NodeImpl *nip);
    e4_VertexImpl *CreateDetachedVertex(const char *nm, int i);
    e4_VertexImpl *CreateDetachedVertex(const char *nm, double d);
    e4_VertexImpl *CreateDetachedVertex(const char *nm, const char *s);
    e4_VertexImpl *CreateDetachedVertex(const char *nm, const void *bytes,
					int nbytes);

    /*
     * Cause deferred node and vertex addition events.
     */

    void CauseDeferredNodeAdditionEvent(e4_NodeImpl *nip);
    void CauseDeferredVertexAdditionEvent(e4_VertexImpl *vip);

    /*
     * The following public methods must be implemented by classes derived
     * from e4_StorageImpl.
     */

    /*
     * Cause a commit *now*.
     */

    virtual bool Commit() = 0;

    /*
     * Is this instance valid?
     */

    virtual bool IsValid() const = 0;

    /*
     * Set the ID of the root node:
     */

    virtual bool SetRootNodeID(int ID) = 0;
    virtual int	 GetRootNodeID() = 0;
    virtual bool IsRootNodeID(int ID) = 0;
    
    /*
     * Given a vertex ID, get the e4_VertexImpl instance representing the
     * next vertex in the containing node.
     */

    virtual e4_VertexImpl *FindNextVertex(int vertexID,
					  e4_VisitMethod vm, int vv,
					  int nameID, 
					  int nodeID,
					  e4_VertexType typeID)
	const = 0;

    /*
     * Given a node ID, get the e4_NodeImpl instance representing the
     * next node in the storage.
     */

    virtual e4_NodeImpl *FindNextNode(int nodeID) const = 0;

    /*
     * Get a value for a statistic being collected for this storage.
     */

    virtual bool GetStatistic(e4_Space sp, e4_SpaceStat st, int &v) = 0;

    /*
     * Return the hash code associated with this storage.
     */

    virtual int HashCode() const = 0;

    /*
     * Is GC deferred?
     */

    virtual bool IsGCDeferred() const = 0;
    virtual bool SetGCDeferral(bool newval) = 0;

    /*
     * Do a GC.
     */

    virtual void DoGC() = 0;

    /*
     * Is a GC needed?
     */

    virtual bool NeedsGC() = 0;

    /*
     * Note that a GC is needed.
     */

    virtual void SetNeedsGC(bool needs) = 0;

protected:

    /*
     * These operations are for use by e4_Storage and e4_StorageVisitor:
     */

    friend class e4_Storage;
    friend class e4_StorageVisitor;

    /*
     * Get the ID of a given node.
     */

    int GetNodeID(e4_Node n) const;

    /*
     * Retrieve the first e4_StorageImpl:
     */

    static e4_StorageImpl *GetFirstStorageImpl();

    /*
     * Retrieve the next e4_StorageImpl after the given one.
     */

    static e4_StorageImpl *GetNextStorageImpl(e4_StorageImpl *si);

    /*
     * This method is called when the reference count goes to (or below)
     * zero.
     */

    virtual void NotReferenced();
    
    /*
     * This method is called to clean up any references from this
     * e4_StorageImpl to other instances.
     */

    void CleanUp();

    /*
     * This method is called to clean up the associated state after a
     * storage is made empty with MakeEmpty().
     */

    void HashCleanup();

    /*
     * The destructor is declared and is virtual so that derived classes can
     * clean up their own state.
     */

    virtual ~e4_StorageImpl();
    
    /*
     * This constructor is used by derived classes.
     */

    e4_StorageImpl(const char *name, const char *dnm);

    /*
     * These operations are for use by e4_NodeImpl and e4_VertexImpl only:
     */

    friend class e4_NodeImpl;
    friend class e4_VertexImpl;
    
    /*
     * Mark the storage as having uncommitted pending changes.
     */

    inline void MarkClean() {dirty = false;};

    /*
     * Find a node given its node ID; create the e4_NodeImpl instance if it is
     * not yet in-memory.
     */

    e4_NodeImpl *FindNode(int nodeID) const;
    e4_NodeImpl *FindOrCreateNode(int nodeID) const;
    e4_NodeImpl *FindOrCreateNode(int nodeID, bool *isNew) const;

    /*
     * Store a node in the activeNodes hash table.
     */

    void StoreNode(int nodeID, e4_NodeImpl *n) const;

    /*
     * Forget a node by removing it from the activeNodes table.
     */

    void ForgetNode(int nodeID);

    /*
     * Retrieve an instance of e4_VertexImpl for a given vertex in a node.
     */

    e4_VertexImpl *GetVertex(int id) const;

    /*
     * Store a vertex in the activeVertices hash table.
     */

    void StoreVertex(int vertexID, e4_VertexImpl *f) const;

    /*
     * Forget a vertex by removing it from the activeVertices table.
     */

    void ForgetVertex(int vertexID);

    /*
     * Is this vertex or node referenced by the user of e4Graph?
     */

    bool IsReferencedNode(int nodeID) const;
    bool IsReferencedVertex(int vertexID) const;

    /*
     * Methods to iterate over all entities referenced by the user program.
     */

    int FirstReferencedNodeID(e4_HashSearch *searchP);
    int NextReferencedNodeID(e4_HashSearch *searchP);
    int FirstReferencedVertexID(e4_HashSearch *searchP);
    int NextReferencedVertexID(e4_HashSearch *searchP);

    /*
     * Enter the name only into the storage-wide hash table. It is already
     * in the persistent storage.
     */

    void AddNameToNameHash(const char *nm, int idx) const;

    /*
     * Given a nameID, get the actual name.
     */

    const char *NameFromNameID(int nameID) const;

    /*
     * Lock the storage -- or unlock it. Returns the old locking status.
     */

    bool SetLocking(bool lockvalue);

    /*
     * Methods below this point are implemented by subclasses
     * of e4_StorageImpl:
     */

    /*
     * Detach a vertex or node.
     */

    virtual bool DetachNodeByID(int id) = 0;
    virtual bool DetachVertexByID(int id) = 0;

    /*
     * Detach the first vertex in parentID that has childID as its node
     * value.
     */

    virtual bool DetachFirstVertexWithNode(int parentID, int childID) = 0;

    /*
     * Determine whether the node or vertex identified by
     * the given ID is detached.
     */

    virtual bool IsDetachedNodeID(int id) = 0;
    virtual bool IsDetachedVertexID(int id) = 0;

    /*
     * Add an element to a free-list.
     */

    virtual void AddToUnusedNodesList(int id) = 0;
    virtual void AddToUnusedIntsList(int id) = 0;
    virtual void AddToUnusedDoublesList(int id) = 0;
    virtual void AddToUnusedStringsList(int id) = 0;
    virtual void AddToUnusedNamesList(int id) = 0;
    virtual void AddToUnusedBinaryList(int id) = 0;

    /*
     * Count how many vertices there are in a node, given the nodeID.
     */

    virtual int VertexCountFromNodeID(int nodeID) const = 0;

    /*
     * Count how many vertices in a node identified by its nodeID have
     * a given name, type or value.
     */

    virtual int VertexCountWithNameFromNodeID(int nodeID, const char *nm) 
	const = 0;
    virtual int VertexCountWithTypeFromNodeID(int nodeID, e4_VertexType tp)
	const = 0;
    virtual int VertexCountWithValueFromNodeID(int nodeID, e4_Value &v)
	const = 0;
    
    /*
     * Find the rank of a child node in the vertex list of a node.
     */

    virtual int GetRankOfChildNode(int parentID,
				   int childID) const = 0;
    
    /*
     * Set a vertex whose vertex ID is given.
     */
    
    virtual bool SetVertex(int vertexID, int nameID, 
			   int vertexType, int itemID) = 0;

    /*
     * Add a vertex according to the rank and io arguments to a node whose
     * node ID is given.
     */
    
    virtual int AddVertex(int nodeID, e4_InsertOrder io, int &rank) = 0;
    
    /*
     * Find the rank of a vertex with a given index in a node whose node ID
     * is given.
     */

    virtual int RankFromVertexID(int nodeID, int vertexID) const = 0;

    /*
     * Find the nth vertex with a given name in a list of vertices.
     */
    
    virtual int VertexIDFromNthVertex(int nodeID, int nameID,
				      int nth, int &rank) const = 0;

    /*
     * Given the rank of a vertex in this node, find the corresponding
     * vertex ID.
     */

    virtual int VertexIDFromRank(int nodeID, int rank) const = 0;

    /*
     * Get the parent node of a node or the parent ID, given the child node ID.
     */

    virtual e4_NodeImpl *GetParentNode(int childID, int nth) const = 0;
    virtual int GetParentNodeID(int childID, int nth) const = 0;

    /*
     * Find the next vertex after this one in the node containing this
     * one.
     */

    virtual e4_VertexImpl *NextVertexAfter(int num, int vertexID) const = 0;

    /*
     * Find the previous vertex before this one in the node containing
     * this vertex.
     */

    virtual e4_VertexImpl *PrevVertexBefore(int num, int vertexID) const = 0;

    /*
     * Get the vertex ID for the vertex containing this node in its parent.
     */

    virtual int GetVertexIDInParent(int parentID, int nodeID) const = 0;

    /*
     * Get vertex IDs for the first or last vertex in a node, given 
     * the node ID.
     */

    virtual int GetFirstVertexID(int nodeID) const = 0;
    virtual int GetLastVertexID(int nodeID) const = 0;

    /*
     * Given a vertex ID, find the e4_NodeImpl instance representing the node
     * that contains that vertex.
     */

    virtual e4_NodeImpl *ContainingNodeFromVertexID(int vertexID) const = 0;

    /*
     * Check that a node ID is legal.
     */

    virtual bool IsLegalNodeID(int index) const = 0;
    
    /*
     * Check that a vertex ID is legal.
     */

    virtual bool IsLegalVertexID(int index) const = 0;

    /*
     * Get the value of a vertex whose vertex ID is given.
     */

    virtual bool GetVertexByIndex(int index, e4_ValueImpl *&v) const = 0;
    virtual bool GetVertexByIndex(int index, e4_NodeImpl *&n) const = 0;
    virtual bool GetVertexByIndex(int index, int &v) const = 0;
    virtual bool GetVertexByIndex(int index, double &f) const = 0;
    virtual bool GetVertexByIndex(int index, const char *&s) const = 0;
    virtual bool GetVertexByIndex(int index, const void *&bytes,
				  int &nbytes) const = 0;

    /*
     * Set the value of a vertex whose vertex ID is given.
     */

    virtual bool SetVertexByIndex(int index, int v) = 0;
    virtual bool SetVertexByIndex(int index, double f) = 0;
    virtual bool SetVertexByIndex(int index, const char *s) = 0;
    virtual bool SetVertexByIndex(int index, const void *bytes, 
				  int nbytes) = 0;
    virtual bool SetVertexByIndexToNode(int index, int childID) = 0;

    /*
     * Rename a vertex whose index is given.
     */

    virtual bool RenameVertexByVertexID(int vertexID, const char *newname) = 0;

    /*
     * Return the vertex type and name of the vertex whose ID is given.
     */

    virtual e4_VertexType VertexTypeFromVertexID(int vertexID) const = 0;
    virtual int NodeIDFromVertexID(int vertexID) const = 0;
    virtual const char *VertexNameFromVertexID(int vertexID) const = 0;
    virtual int NameIDFromVertexID(int vertexID) const = 0;

    /*
     * Retrieve the value of an entry. If the vertex identified by the
     * given index contains a value of the requested type, the value is
     * placed in the output parameter and true is returned. Otherwise,
     * false is returned and the value of the output parameter is left
     * unmodified.
     */

    virtual bool GetNode(int index, e4_NodeImpl *&value) const = 0;
    virtual bool GetInt(int index, int &value) const = 0;
    virtual bool GetDouble(int index, double &value) const = 0;
    virtual bool GetString(int index, char *&value) const = 0;
    virtual bool GetBinary(int index, void *&value, int &nbytes) const = 0;
    virtual bool GetVertexName(int index, const char *&value) const = 0;

    /*
     * Specialized version of GetNode that also returns an indication
     * whether the node is newly created.
     */

    virtual bool GetNode(int index, e4_NodeImpl *&value, bool *isNew) 
	const = 0;
    
    /*
     * Add a new entry to one of the type-specific spaces. The operation
     * returns the index of the new item.
     */

    virtual int ReserveNodeID() = 0;
    virtual int ReserveVertexID(const char *nm) = 0;
    virtual int AddInt(int value) = 0;
    virtual int AddDouble(double value) = 0;
    virtual int AddString(const char *value) = 0;
    virtual int AddBinary(const void *value, int nbytes) = 0;
    virtual int AddName(const char *nm) = 0;

    /*
     * Add the node identified by "parentID" to the list of nodes
     * that contain the node identified by "childID".
     */

    virtual void AddParent(int childID, int parentID, int vertexID) = 0;

    /*
     * Move the vertex identified by vertexID into the node identified by
     * nodeID at the given rank.
     */

    virtual bool MoveVertex(int vertexID, int nodeID, int rank) = 0;

    /*
     * Use driver-specific methods to copy from one storage to another.
     */

    virtual bool CopyTo(e4_StorageImpl *osp) = 0;

    /*
     * Use driver-specific methods to destroy the physical storage.
     */

    virtual void Destroy() = 0;

    /*
     * How many nodes in this storage are parents of the node with the
     * given ID?
     */

    virtual int ParentCount(int nodeID) const = 0;

    /*
     * How many times does the node with the given ID appear as the
     * value of a vertex in this storage?
     */

    virtual int OccurrenceCount(int nodeID) const = 0;

    /*
     * How many times does the node with the given child ID appear as
     * the value of a vertex in the node with the given parent ID?
     */

    virtual int OccurrenceCount(int nodeID, int parentID) const = 0;

    /*
     * What is the rank of the node with the given parent ID for the
     * node with the given child ID?
     */

    virtual int ParentRank(int nodeID, int parentID) const = 0;

    /*
     * User data manipulation facility:
     */

    virtual bool GetNodeUserData(int nodeID, int &userData) const = 0;
    virtual bool GetVertexUserData(int vertexID, int &userData) const = 0;

    virtual bool SetNodeUserData(int nodeID, int userData) const = 0;
    virtual bool SetVertexUserData(int vertexID, int userData) const = 0;
};

/*
 * The e4_NodeImpl class:
 */

class e4_DLL e4_NodeImpl : public e4_RefCounter {
private:
    e4_NodeImpl();				/* Ensure that noone can
						 * use the default
						 * constructor. */
    e4_NodeImpl(const e4_NodeImpl &);		/* Nor copy construction */
    void operator= (const e4_NodeImpl *);	/* Nor assignment */

private:
    e4_StorageImpl *storage;			/* In which storage is this
                                                 * node? */
    int nodeID;					/* The ID for this node. */

public:
    /*
     * There is no public constructor. The only way to get an 
     * e4_NodeImpl instance is to call an API on a storage or
     * another entity that returns a node.
     */

    ~e4_NodeImpl();				/* Deletes the e4_NodeImpl *BUT
						 * NOT* the datum. This
						 * just releases the program
						 * memory data structure which
						 * is only a cache for the
						 * node in its view. Use
						 * e4_NodeImpl::Detach() 
						 * to detach a node. */

    /*
     * How many vertices are currently in this node?
     */

    int VertexCount() const;

    /*
     * How many vertices in this node have the given name, type or value?
     */

    int VertexCountWithName(const char *nm) const;
    int VertexCountWithType(e4_VertexType tp) const;
    int VertexCountWithValue(e4_Value &v) const;

    /*
     * Set or add a vertex whose value is a node:
     */

    e4_NodeImpl *SetNthNode(const char *nm, int nth);
    e4_NodeImpl *SetNode(const char *nm);
    e4_NodeImpl *SetNodeByRank(int rank);
    e4_NodeImpl *AddNode(const char *nm, e4_InsertOrder order, int &rank);

    /*
     * Operations to set data into existing vertices in this node.
     */

    bool SetNthVertex(const char *nm, int nth, int v);
    bool SetNthVertex(const char *nm, int nth, double f);
    bool SetNthVertex(const char *nm, int nth, const char *s);
    bool SetNthVertex(const char *nm, int nth, const void *b, int nbytes);
    bool SetNthVertexToNode(const char *nm, int nth, int childID);

    /*
     * Set the value of a vertex identified by rank. The return value is
     * true if the vertex is new and false if it existed beforehand.
     */
    
    bool SetVertexByRank(int rank, int v);
    bool SetVertexByRank(int rank, double f);
    bool SetVertexByRank(int rank, const char *s);
    bool SetVertexByRank(int rank, const void *bytes, int nbytes);
    bool SetVertexByRankToNode(int rank, int childID);

    /*
     * These operations create a new vertex with the requested name and
     * add it at the position specified by the order and rank arguments.
     * The rank argument is in-out; on the way in, in combination with
     * the order argument, it determines where in the node the new vertex
     * is created. On the way out, it gives the caller the rank of the
     * new vertex created. Upon success, true is returned. If an error
     * occurred, rank is set to E4_VERTEXNOTCREATED and false is returned.
     */

    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank, int v);
    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank, double f);
    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank,
		   const char *s);
    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank,
		   const void *b, int nbytes);
    bool AddVertexWithNode(const char *nm, e4_InsertOrder order, int &rank,
			   int childID);

    /*
     * Operations to add a vertex and return an e4_VertexImpl for the 
     * new vertex in one step. This is much more efficient than AddVertex
     * followed by GetVertexRef.
     */

    e4_VertexImpl *AddVertexRef(const char *nm, e4_InsertOrder order, 
				int &rank, int v);
    e4_VertexImpl *AddVertexRef(const char *nm, e4_InsertOrder order,
				int &rank, double f);
    e4_VertexImpl *AddVertexRef(const char *nm, e4_InsertOrder order,
				int &rank, const char *s);
    e4_VertexImpl *AddVertexRef(const char *nm, e4_InsertOrder order,
				int &rank, const void *b, int nbytes);
    e4_VertexImpl *AddVertexRefWithNode(const char *nm, e4_InsertOrder order,
					int &rank, int childID);

    e4_VertexImpl *AddNodeRef(const char *nm, e4_InsertOrder order, int &rank,
			      e4_NodeImpl *&n);

    /*
     * Operations to get data from vertices in this node. These operations
     * return true if the vertex is found and then the output parameter is
     * modified to refer to the vertex value. If the vertex is not found then
     * the operation returns false and the output parameter is not modified.
     *
     * The char * and void * returned are valid only until the next call to
     * any e4Tree API. If the caller needs to keep access to the value,
     * copy the value into memory owned by the caller.
     */

    bool GetNthVertex(const char *nm, int nth, e4_ValueImpl *&v) const;
    bool GetNthVertex(const char *nm, int nth, e4_NodeImpl *&n) const;
    bool GetNthVertex(const char *nm, int nth, int &v) const;
    bool GetNthVertex(const char *nm, int nth, double &f) const;
    bool GetNthVertex(const char *nm, int nth, const char *&s) const;
    bool GetNthVertex(const char *nm, int nth,
		      const void *&v, int &nbytes) const;

    bool GetVertexByRank(int rank, e4_ValueImpl *&v) const;
    bool GetVertexByRank(int rank, e4_NodeImpl *&n) const;
    bool GetVertexByRank(int rank, int &v) const;
    bool GetVertexByRank(int rank, double &f) const;
    bool GetVertexByRank(int rank, const char *&s) const;
    bool GetVertexByRank(int rank, const void *&v, int &nbytes) const;

    /*
     * Detach a vertex:
     */

    bool DetachVertex(const char *nm, int nth);
    bool DetachVertexByRank(int rank);

    /*
     * Detach the first vertex whose value is a given child node.
     */

    bool DetachFirstVertexWithNode(e4_NodeImpl *childImpl);

    /*
     * Get an e4_VertexImpl pointer for a vertex identified either by rank or
     * by name and nth.
     */

    e4_VertexImpl *GetVertexRef(const char *nm, int nth);
    e4_VertexImpl *GetVertexRefByRank(int rank);

    /*
     * This operation returns the type of the datum stored in the
     * named vertex. If the vertex does not exist then E4_VERTEXNOTFOUND
     * is returned.
     */

    e4_VertexType VertexType(const char *nm, int nth) const;
    e4_VertexType VertexTypeByRank(int rank) const;

    /*
     * Return the name of the vertex whose rank is given. Returns NULL
     * if no such vertex is present.
     */

    const char *VertexName(int rank) const;

    /*
     * Rename a vertex identified by a rank. Returns true if the vertex
     * exists, false otherwise.
     */

    bool RenameVertex(int rank, const char *newNm);

    /*
     * This operation retrieves the rank of the "n"th vertex with the given
     * name in this node. The operation returns E4_VERTEXNOTFOUND if there is
     * no matching vertex, or the rank of the matching vertex.
     */

    int VertexRank(const char *nm, int nth) const;
    bool Exists(const char *nm, int nth) const;

    /*
     * Node->Node operations:
     */

    e4_NodeImpl *GetParent(int nth) const;

    /*
     * How many parents does this node have?
     */

    int ParentCount() const;

    /*
     * In how many different vertices is this node the vertex target?
     */

    int OccurrenceCount() const;

    /*
     * How many vertices point from this parent to this child node?
     */

    int OccurrenceCount(int parentID) const;

    /*
     * What is the rank of the given parent for this node?
     */

    int ParentRank(int parentID) const;

    /*
     * Node->Storage: get the storage in which this node appears.
     */

    e4_StorageImpl *GetStorage() const;
    
    /*
     * Return the rank of this node in its parent. If this node is a root
     * then it returns E4_ROOTINDEX, else it returns the rank of the vertex
     * whose datum is this node, in its parent.
     */

    int GetRankInParent(int nth) const;

    /*
     * A node is the "root" if it is the node returned by s.GetRootNode()
     * when called on its storage.
     */

    bool IsRoot() const;

    /*
     * Return the name of this node in its parent. If this node is a root
     * then it returns NULL, else it returns the name of the vertex whose
     * datum is this node, in its parent.
     */

    const char *GetNameInParent(int nth) const;

    /*
     * This operation returns an ID which uniquely identifies this e4_NodeImpl
     * in the e4_StorageImpl that contains it.
     */

    int GetUniqueID() const;

    /*
     * Detach this node from all vertices for which it is the value.
     */

    bool Detach() const;

    /*
     * Is this node detached?
     */

    bool IsDetached() const;

    /*
     * Is this instance valid?
     */

    virtual bool IsValid() const;

    /*
     * Manage the user data associated with this node.
     */

    bool GetUserData(int &userData) const;
    bool SetUserData(int userData) const;

    /*
     * Move the vertex identified by vertexID into the node identified
     * by nodeID at the given rank.
     */

    bool MoveVertex(int vertexID, e4_InsertOrder order, int rank);

    /*
     * Cause a deferred E4_CBADDNODE event.
     */

    void CauseDeferredNodeAdditionEvent();

protected:

    /*
     * This method is called when the reference count goes to (or below)
     * zero.
     */

    virtual void NotReferenced();

    /*
     * These methods are for use by e4_StorageImpl and e4_VertexImpl only:
     */

    friend class e4_StorageImpl;
    friend class e4_VertexImpl;

    /*
     * Constructor which takes a node ID:
     */

    e4_NodeImpl(e4_StorageImpl *s, int nodeID);

    /*
     * Get the index of the first and last vertex in this node.
     */

    int GetFirstVertexID() const;
    int GetLastVertexID() const;

    /*
     * Get the node ID.
     */

    int GetNodeID() const;

    /*
     * Set the storage instance variable.
     */

    void SetStorage(e4_StorageImpl *newstorage);
};

/*
 * The e4_VertexImpl class:
 */

class e4_DLL e4_VertexImpl : public e4_RefCounter {
private:
    e4_VertexImpl();				/* Ensure that noone can
						 * use the default
						 * constructor. */
    e4_VertexImpl(const e4_VertexImpl &);	/* Nor copy construction. */
    void operator= (const e4_VertexImpl &);	/* Nor assignment. */

private:
    e4_StorageImpl *s;				/* The e4_StorageImpl that the
                                                 * containing vertex appears
                                                 * in. */
    int vertexID;				/* Unique ID for this 
						 * e4_VertexImpl. */
public:
    /*
     * Destructor:
     */

    ~e4_VertexImpl();				/* Deletes the e4_VertexImpl
						 * BUT NOT the datum. This just
						 * releases the program
						 * memory data structure which
						 * is only a cache for the
						 * view-based vertex. To remove
						 * the underlying vertex, use
						 * e4_VertexImpl::Detach and
						 * invalidate all references
						 * to it. Storage is recycled
						 * automatically. */

    /*
     * Get the vertex content.
     */

    inline bool Get(e4_ValueImpl *&v) const {
        if (s == NULL) {
            return false;
        }
	return s->GetVertexByIndex(vertexID, v);
    }
    inline bool Get(e4_NodeImpl *&node) const {
        if (s == NULL) {
            return false;
        }
	return s->GetVertexByIndex(vertexID, node);
    }
    inline bool Get(int &v) const {
	if (s == NULL) {
	    return false;
	}
	return s->GetVertexByIndex(vertexID, v);
    }
    inline bool Get(double &fv) const {
	if (s == NULL) {
	    return false;
	}
	return s->GetVertexByIndex(vertexID, fv);
    }
    inline bool Get(const char *&str) const {
	if (s == NULL) {
	    return false;
	}
	return s->GetVertexByIndex(vertexID, str);
    }
    inline bool Get(const void *&bytes, int &nbytes) const {
	if (s == NULL) {
	    return false;
	}
	return s->GetVertexByIndex(vertexID, bytes, nbytes);
    }

    /*
     * Set the target (value) of the vertex (and potentially its type).
     */

    inline bool Set(int v) {
	if (s == NULL) {
	    return false;
	}
	if (s->IsLocked()) {
	    return false;
	}
	if (!s->SetVertexByIndex(vertexID, v)) {
	    return false;
	}
	if (s->HasCallbacks(E4_CBMODVERTEX)) {
	    s->CallbackEvent(E4_CBMODVERTEX, this);
	}
	return true;
    }
    inline bool Set(double fv) {
	if (s == NULL) {
	    return false;
	}
	if (!s->SetVertexByIndex(vertexID, fv)) {
	    return false;
	}
	if (s->HasCallbacks(E4_CBMODVERTEX)) {
	    s->CallbackEvent(E4_CBMODVERTEX, this);
	}
	return true;
    }
    inline bool Set(const char *str) {
	if (s == NULL) {
	    return false;
	}
	if (s->IsLocked()) {
	    return false;
	}
	if (!s->SetVertexByIndex(vertexID, str)) {
	    return false;
	}
	if (s->HasCallbacks(E4_CBMODVERTEX)) {
	    s->CallbackEvent(E4_CBMODVERTEX, this);
	}
	return true;
    }
    inline bool Set(const void *bytes, int nbytes) {
	if (s == NULL) {
	    return false;
	}
	if (s->IsLocked()) {
	    return false;
	}
	if (!s->SetVertexByIndex(vertexID, bytes, nbytes)) {
	    return false;
	}
	if (s->HasCallbacks(E4_CBMODVERTEX)) {
	    s->CallbackEvent(E4_CBMODVERTEX, this);
	}
	return true;
    }
    bool SetToNode(int childID);

    /*
     * Set the value (target) of this vertex to a new node.
     */

    e4_NodeImpl *SetNode() const;

    /*
     * Get the rank of this vertex in its source node.
     */

    inline int Rank() const {
        int nodeID;
        
	if (s == NULL) {
	    return E4_VERTEXNOTFOUND;
	}
        nodeID = s->NodeIDFromVertexID(vertexID);
        if (nodeID == E4_NODENOTFOUND) {
            return E4_VERTEXNOTFOUND;
        }
        return s->RankFromVertexID(nodeID, vertexID);
    }

    /*
     * Return the next vertex after this in this node.
     */

    inline e4_VertexImpl *Next(int num) const
    {
	if (s == NULL) {
	    return NULL;
	}
	return s->NextVertexAfter(num, vertexID);
    }

    /*
     * Return the previous vertex before this in this node.
     */

    inline e4_VertexImpl *Prev(int num) const
    {
	if (s == NULL) {
	    return NULL;
	}
	return s->PrevVertexBefore(num, vertexID);
    }

    /*
     * Detach this vertex from its containing node.
     */

    bool Detach() const;

    /*
     * Is this vertex detached?
     */

    bool IsDetached() const;

    /*
     * Is this vertex valid?
     */

    inline virtual bool IsValid() const {
	if (s == NULL) {
	    return false;
	}
	return (s->IsValid() && s->IsLegalVertexID(vertexID));
    }

    /*
     * Get the vertex type for this vertex.
     */

    inline e4_VertexType Type() const {
	if (s == NULL) {
	   return E4_VTUNKNOWN;
	}
	return s->VertexTypeFromVertexID(vertexID);
    }

    /*
     * Get the name of this vertex.
     */

    inline const char *Name() const {
	if (s == NULL) {
	    return NULL;
	}
	return s->VertexNameFromVertexID(vertexID);
    }

    /*
     * Rename this vertex.
     */

    inline bool Rename(const char *newname) const {
	if ((s == NULL) || (newname == NULL)) {
	    return false;
	}
	if (s->IsLocked()) {
	    return false;
	}
	if (!s->RenameVertexByVertexID(vertexID, newname)) {
	    return false;
	}
	if (s->HasCallbacks(E4_CBMODVERTEX)) {
	    s->CallbackEvent(E4_CBMODVERTEX, (e4_RefCounter *) this);
	}
	return true;
    }
	
    /*
     * This operation returns an ID which uniquely identifies this 
     * e4_VertexImpl within the e4_StorageImpl that contains it.
     */

    inline int GetUniqueID() const {
	if (s == NULL) {
	    return E4_VERTEXNOTFOUND;
	}
	return vertexID;
    }

    /*
     * Get the storage and node back out of this vertex.
     */

    inline e4_StorageImpl *GetStorage() const {return s;}
    inline e4_NodeImpl *GetNode() const {
        if (s == NULL) {
            return NULL;
        }
        return s->ContainingNodeFromVertexID(vertexID);
    }

    /*
     * Move the vertex identified by moveVertexID to the given position
     * within the node containing this vertex.
     */

    inline bool MoveVertex(int moveVertexID, int rank) const {
	int nodeID;

	if (s == NULL) {
	    return false;
	}
	if (s->IsLocked()) {
	    return false;
	}
	nodeID = s->NodeIDFromVertexID(vertexID);
	if (nodeID == E4_NODENOTFOUND) {
	    return false;
	}
	if (!s->MoveVertex(moveVertexID, nodeID, rank)) {
	    return false;
	}
	if (s->HasCallbacks(E4_CBMODVERTEX)) {
	    s->CallbackEvent(E4_CBMODVERTEX, s->GetVertex(moveVertexID));
	}
	return true;
    }

    /*
     * Manage the user data associated with this vertex.
     */

    bool GetUserData(int &userData) const;
    bool SetUserData(int userData) const;

protected:

    /*
     * This method is called when the reference count goes to (or below)
     * zero.
     */

    virtual void NotReferenced();

    /*
     * These operations are for use by e4_StorageImpl only:
     */

    friend class e4_StorageImpl;

    /*
     * Constructor:
     */

    e4_VertexImpl(e4_StorageImpl *st, int ii);

    /*
     * Set the storage instance variable of this vertex.
     */

    inline void SetStorage(e4_StorageImpl *ns) {s = ns;}
};
	
#endif	/* __E4_GRAPHIMPL_H__ */
