/*
 * e4storage.cpp --
 *
 *	Implementation of the e4_Storage class defined in e4graph.h.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "e4graph.h"
#include "e4graphimpl.h"

/*
 * This instance of e4_Storage can be used (by assigning it to fields or
 * variables of type e4_Storage) to remove remaining references to instances
 * that you want to discard. Especially useful for removing remaining
 * references in malloc'ed memory.
 */

e4_Storage const invalidStorage;

/*
 * Default constructor:
 */

e4_Storage::e4_Storage() : e4_RefCount() {}

/*
 * Constructor which assigns a value to the the implementation pointer.
 */

e4_Storage::e4_Storage(e4_StorageImpl *ip) : e4_RefCount(ip) {}

/*
 * Copying constructor:
 */

e4_Storage::e4_Storage(const e4_Storage &referrer) : e4_RefCount(referrer) {}

/*
 * Assignment operator:
 */

e4_Storage & e4_Storage::operator=(const e4_Storage &referrer)
{
    return (e4_Storage &) e4_RefCount::operator=(referrer);
}

/*
 * Constructor that chooses a concrete representation for the underlying
 * implementation.
 */

e4_Storage::e4_Storage(const char *nm, const char *drivername)
    : e4_RefCount()
{
    impl = e4_StorageImpl::GetStorage(nm, drivername);
    if (impl != NULL) {
	impl->IncrRefCount();
    }
}

/*
 * Turn auto-commit on or off.
 */

void
e4_Storage::AutoCommit(bool on) const
{
    if (impl != NULL) {
	((e4_StorageImpl *) impl)->AutoCommit(on);
    }
}

/*
 * Is auto-commit turned on?
 */

bool
e4_Storage::AutoCommit() const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->AutoCommit();
}

/*
 * Commit the storage now.
 */

bool
e4_Storage::Commit() const
{
    if (impl != NULL) {
	return ((e4_StorageImpl *) impl)->Commit();
    }
    return false;
}

/*
 * Copy this storage to another storage.
 */

bool
e4_Storage::CopyTo(e4_Storage otherStorage, bool forceCommit) const
{
    if (!IsValid() || !otherStorage.IsValid()) {
	return false;
    }
    if (!((e4_StorageImpl *) impl)->
		CopyTo((e4_StorageImpl *) otherStorage.impl)) {
	return false;
    }
    if (forceCommit) {
	return otherStorage.Commit();
    }
    return true;
}

/*
 * Destroy this storage. If auto-commit is set, first commit.
 */

bool
e4_Storage::Delete()
{
    if (impl != NULL) {
	if (((e4_StorageImpl *) impl)->AutoCommit()) {
	    ((e4_StorageImpl *) impl)->Commit();
	}
	((e4_StorageImpl *) impl)->Destroy();
	impl->DecrRefCount();
    }
    impl = NULL;

    return true;
}

/*
 * Get and set the root node.
 */

bool
e4_Storage::GetRootNode(e4_Node &n) const
{
    e4_NodeImpl *nnip;

    if (impl == NULL) {
	return false;
    }

    nnip = ((e4_StorageImpl *) impl)->GetRootNode();
    if ((nnip == NULL) || (!nnip->IsValid())) {
	return false;
    }

    e4_Node nn(nnip);

    n = nn;

    return true;
}

bool
e4_Storage::SetRootNode(e4_Node n) const
{
    e4_Storage hisstorage;

    if (!n.IsValid() || (impl == NULL) || (!n.GetStorage(hisstorage)) ||
	(*this != hisstorage)) {
	return false;
    }

    ((e4_StorageImpl *) impl)->SetRootNode((e4_NodeImpl *) n.impl);

    return true;
}

/*
 * Given a unique ID, obtain the referenced node.
 */

bool
e4_Storage::GetNodeFromID(e4_NodeUniqueID nid, e4_Node &n) const
{
    e4_NodeImpl *nnip;

    if (impl == NULL) {
	return false;
    }
    nnip = ((e4_StorageImpl *) impl)->FindNode(nid.GetUniqueID());
    if (nnip == NULL) {
	return false;
    }
    e4_Node nn(nnip);

    n = nn;

    return true;
}

/*
 * Given a unique ID, obtain the vertex referenced by the ID.
 */

bool
e4_Storage::GetVertexFromID(e4_VertexUniqueID vid, e4_Vertex &v) const
{
    e4_VertexImpl *vvip;

    if (impl == NULL) {
	return false;
    }
    vvip = ((e4_StorageImpl *) impl)->GetVertex(vid.GetUniqueID());
    if (vvip == NULL) {
	return false;
    }

    e4_Vertex vv(vvip);

    v = vv;

    return true;
}

/*
 * Given a unique ID for a vertex, get the next vertex after it in the same
 * node.
 */

bool
e4_Storage::FindNextVertex(int vertexID, e4_VisitMethod vm, int vf,
			   int nameID, int nodeID,
			   e4_VertexType typeID, e4_Vertex &f) const
{
    e4_VertexImpl *fi;

    if (impl == NULL) {
	return false;
    }
    fi = ((e4_StorageImpl *) impl)->FindNextVertex(vertexID, vm, vf,
						   nameID, nodeID, typeID);
    if (fi == NULL) {
	return false;
    }
    e4_Vertex ff(fi);

    f = ff;

    return true;
}

/*
 * Given a unique ID for a node, find the next one after it in
 * the storage.
 */

bool
e4_Storage::FindNextNode(int nodeID, e4_Node &n) const
{
    e4_NodeImpl *ni;

    if (impl == NULL) {
	return false;
    }
    ni = ((e4_StorageImpl *) impl)->FindNextNode(nodeID);
    if (ni == NULL) {
	return false;
    }

    e4_Node nn(ni);

    n = nn;

    return true;
}

/*
 * Is this storage dirty?
 */

bool
e4_Storage::IsDirty() const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->IsDirty();
}

void
e4_Storage::MarkDirty() const
{
    if (impl != NULL) {
	((e4_StorageImpl *) impl)->MarkDirty();
    }
}

/*
 * Is this storage locked?
 */

bool
e4_Storage::IsLocked() const
{
    if (impl == NULL) {
	return true;
    }
    return ((e4_StorageImpl *) impl)->IsLocked();
}

/*
 * Get the name of this storage.
 */

const char *
e4_Storage::GetName() const
{
    if (impl == NULL) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->GetName();
}

/*
 * Get the ID of the storage driver used by this storage.
 */

const char *
e4_Storage::GetDriver() const
{
    if (impl == NULL) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->GetDriver();
}

bool
e4_Storage::CreateDetachedNode(e4_Node &n) const
{
    e4_NodeImpl *nip;

    if (impl == NULL) {
	return false;
    }
    nip = ((e4_StorageImpl *) impl)->CreateDetachedNode();
    if (nip == NULL) {
	return false;
    }
    
    e4_Node nn(nip);

    n = nn;

    ((e4_StorageImpl *) impl)->CauseDeferredNodeAdditionEvent(nip);

    return true;
}

bool
e4_Storage::CreateDetachedVertex(const char *nm, e4_Node n,
				 e4_Vertex &v) const
{
    e4_VertexImpl *vip;

    if (impl == NULL) {
	return false;
    }
    vip = ((e4_StorageImpl *) impl)->CreateDetachedVertex(nm, 
	(e4_NodeImpl *) n.impl);
    if (vip == NULL) {
	return false;
    }

    e4_Vertex vv(vip);

    v = vv;

    ((e4_StorageImpl *) impl)->CauseDeferredVertexAdditionEvent(vip);

    return true;
}

bool
e4_Storage::CreateDetachedVertex(const char *nm, int i, e4_Vertex &v) const
{
    e4_VertexImpl *vip;

    if (impl == NULL) {
	return false;
    }
    vip = ((e4_StorageImpl *) impl)->CreateDetachedVertex(nm, i);
    if (vip == NULL) {
	return false;
    }

    e4_Vertex vv(vip);

    v = vv;

    ((e4_StorageImpl *) impl)->CauseDeferredVertexAdditionEvent(vip);

    return true;
}

bool
e4_Storage::CreateDetachedVertex(const char *nm, double d, e4_Vertex &v)
	const
{
    e4_VertexImpl *vip;

    if (impl == NULL) {
	return false;
    }
    vip = ((e4_StorageImpl *) impl)->CreateDetachedVertex(nm, d);
    if (vip == NULL) {
	return false;
    }

    e4_Vertex vv(vip);

    v = vv;

    ((e4_StorageImpl *) impl)->CauseDeferredVertexAdditionEvent(vip);

    return true;
}

bool
e4_Storage::CreateDetachedVertex(const char *nm, const char *s, e4_Vertex &v)
	const
{
    e4_VertexImpl *vip;

    if (impl == NULL) {
	return false;
    }
    vip = ((e4_StorageImpl *) impl)->CreateDetachedVertex(nm, s);
    if (vip == NULL) {
	return false;
    }
    
    e4_Vertex vv(vip);

    v = vv;

    ((e4_StorageImpl *) impl)->CauseDeferredVertexAdditionEvent(vip);

    return true;
}

bool
e4_Storage::CreateDetachedVertex(const char *nm, const void *b, int nb,
				 e4_Vertex &v) const
{
    e4_VertexImpl *vip;

    if (impl == NULL) {
	return false;
    }
    vip = ((e4_StorageImpl *) impl)->CreateDetachedVertex(nm, b, nb);
    if (vip == NULL) {
	return false;
    }

    e4_Vertex vv(vip);

    v = vv;

    ((e4_StorageImpl *) impl)->CauseDeferredVertexAdditionEvent(vip);

    return true;
}

bool
e4_Storage::CreateDetachedVertex(const char *nm, e4_Value &vv, e4_Vertex &v)
	const
{
    if (impl == NULL) {
	return false;
    }
    switch (vv.vertexType) {
    case E4_VTNODE:
	return CreateDetachedVertex(nm, vv.n, v);
    case E4_VTINT:
	return CreateDetachedVertex(nm, vv.u.i, v);
    case E4_VTDOUBLE:
	return CreateDetachedVertex(nm, vv.u.d, v);
    case E4_VTSTRING:
	return CreateDetachedVertex(nm, vv.u.s, v);
    case E4_VTBINARY:
	return CreateDetachedVertex(nm, vv.u.b.bytes, vv.u.b.nbytes, v);
    default:
	return false;
    }
}

/*
 * Get a value for a statistic being collected for this storage.
 */

bool
e4_Storage::GetStatistic(e4_Space sp, e4_SpaceStat st, int &v) const
{
    if (impl == NULL) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->GetStatistic(sp, st, v);
}

/*
 * Intern a name.
 */

int
e4_Storage::InternName(const char *nm) const
{
    if (impl == NULL) {
	return E4_VERTEXNOTFOUND;
    }
    return ((e4_StorageImpl *) impl)->InternName(nm);
}

/*
 * Get a node given its node ID:
 */

e4_NodeImpl *
e4_Storage::GetNode(int nodeID) const
{
    if (!IsValid()) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->FindOrCreateNode(nodeID);
}

/*
 * Get the name of a vertex given the ID of the name.
 */

const char *
e4_Storage::GetName(int nameID) const
{
    if (!IsValid()) {
	return NULL;
    }
    return ((e4_StorageImpl *) impl)->NameFromNameID(nameID);
}

/*
 * Return the storage implementation for this storage:
 */

e4_StorageImpl *
e4_Storage::GetStorageImpl() const
{
    if (!IsValid()) {
	return NULL;
    }
    return (e4_StorageImpl *) impl;
}

/*
 * Is GC deferred?
 */

bool
e4_Storage::IsGCDeferred() const
{
    if (!IsValid()) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->IsGCDeferred();
}

bool
e4_Storage::SetGCDeferral(bool newval) const
{
    if (!IsValid()) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->SetGCDeferral(newval);
}

/*
 * Do a GC.
 */

void
e4_Storage::DoGC() const
{
    if (IsValid()) {
	((e4_StorageImpl *) impl)->DoGC();
    }
}

/*
 * Is a GC needed?
 */

bool
e4_Storage::NeedsGC() const
{
    if (!IsValid()) {
	return false;
    }
    return ((e4_StorageImpl *) impl)->NeedsGC();
}

/*
 * Callback facility: Declare or remove callbacks for specific kinds of
 * events.
 */

bool
e4_Storage::DeclareNodeAddCallback(e4_NodeAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBADDNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteNodeAddCallback(e4_NodeAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBADDNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareVertexAddCallback(e4_VertexAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBADDVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteVertexAddCallback(e4_VertexAddCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBADDVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareNodeDetCallback(e4_NodeDetCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBDETNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteNodeDetCallback(e4_NodeDetCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBDETNODE, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareVertexDetCallback(e4_VertexDetCallbackFunction fn, 
				     void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBDETVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteVertexDetCallback(e4_VertexDetCallbackFunction fn,
				    void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBDETVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeclareVertexModCallback(e4_VertexModCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->AddCallback(E4_CBMODVERTEX, (void *) fn, cd);
    return true;
}

bool
e4_Storage::DeleteVertexModCallback(e4_VertexModCallbackFunction fn, void *cd)
{
    if (!IsValid()) {
	return false;
    }
    ((e4_StorageImpl *) impl)->DelCallback(E4_CBMODVERTEX, (void *) fn, cd);
    return true;
}

/*
 * Return the kind identifier for this instance.
 */

e4_RefKind
e4_Storage::Kind() const
{
    return E4_RKSTORAGE;
}
